"""Buildbot PBChangeSource hook for bzr.
"""

# Options from the svn hook that might be nice to have in the future:
#        ['include', 'f', None,
#         '''\
#Search the list of changed files for this regular expression, and if there is
#at least one match notify buildbot; otherwise buildbot will not do a build.
#You may provide more than one -f argument to try multiple
#patterns.  If no filter is given, buildbot will always be notified.'''],
#        ['exclude', 'F', None,
#         '''\
#The inverse of --include.  Changed files matching this expression will never
#be considered for a build.
#You may provide more than one -F argument to try multiple
#patterns.  Excludes override includes, that is, patterns that match both an
#include and an exclude will be excluded.'''],
#        ]

#import subprocess # used for sendchange experiment, also commented out below.

import bzrlib.errors
import twisted.internet.reactor
import twisted.spread.pb
import twisted.cred.credentials

import bzrbuildbot.change

HOOK_KEY = 'buildbot_on'
SERVER_KEY = 'buildbot_server'
PORT_KEY = 'buildbot_port'
DRYRUN_KEY = 'buildbot_dry_run'
PQM_KEY = 'buildbot_pqm'
#SENDCHANGE_KEY = 'buildbot_use_sendchange'
SEND_BRANCHNAME_KEY = 'buildbot_send_branch_name'

PUSH_VALUE = 'push'
COMMIT_VALUE = 'commit'
CHANGE_VALUE = 'change'

def _is_true(config, key):
    val = config.get_user_option(key)
    return val is not None and val.lower().strip() in (
        'y', 'yes', 't', 'true')

def _installed_hook(branch):
    value = branch.get_config().get_user_option(HOOK_KEY)
    if value is not None:
        value = value.strip().lower()
        if value not in (PUSH_VALUE, COMMIT_VALUE, CHANGE_VALUE):
            raise bzrlib.errors.BzrError(
                '%s, if set, must be one of %s, %s, or %s' % (
                    HOOK_KEY, PUSH_VALUE, COMMIT_VALUE, CHANGE_VALUE))
    return value

def send_change(branch, old_revno, old_revid, new_revno, new_revid, hook):
    config = branch.get_config()
    server = config.get_user_option(SERVER_KEY)
    if not server:
        raise bzrlib.errors.BzrError(
            'If %s is set, %s must be set' % (HOOK_KEY, SERVER_KEY))
    change = bzrbuildbot.change.generate(
        branch, old_revno, old_revid, new_revno, new_revid,
        blame_merge_author=_is_true(config, PQM_KEY))
    if _is_true(config, SEND_BRANCHNAME_KEY):
        change['branch'] = branch.nick
    # as of this writing (in Buildbot 0.7.9), 9989 is the default port when
    # you make a buildbot master.
    port = int(config.get_user_option(PORT_KEY) or 9989)
    # if dry run, stop.
    if _is_true(config, DRYRUN_KEY):
        msg = "DRY RUN (*not* sending changes to %s:%d on %s)" % (
            server, port, hook)
        print " bzrbuildout:", msg
        keys = change.keys()
        keys.sort()
        for k in keys:
            print "[%10s]: %s" % (k, change[k])
        return
    #if _is_true(config, SENDCHANGE_KEY):
    #    msg = "SENDCHANGE to buildbot master %s:%d on %s" % (
    #        server, port, hook)
    #    print " bzrbuildbot:", msg
    #    cmd = ['buildbot', 'sendchange',
    #           '--master', ':'.join((server, str(port))),
    #           '--username', change['who'],
    #           '--revision_number', str(change['revision']),
    #           '--comments', change['comments']]
    #    if 'branch' in change:
    #        cmd.extend(['--branch', change['branch']])
    #    cmd.extend(files)
    #    subprocess.call(cmd)
    #    return
    pbcf = twisted.spread.pb.PBClientFactory()
    twisted.internet.reactor.connectTCP(server, port, pbcf)
    deferred = pbcf.login(
        twisted.cred.credentials.UsernamePassword('change', 'changepw'))

    def sendChanges(remote):
        """Send changes to buildbot."""
        msg = "sending changes: %s"  % change
        print " bzrbuildout:", msg
        return remote.callRemote('addChange', change)

    deferred.addCallback(sendChanges)

    def quit(ignore, msg):
        print " bzrbuildout:", msg
        twisted.internet.reactor.stop()

    def failed(failure):
        print " bzrbuildout: FAILURE"
        print failure
        twisted.internet.reactor.stop()

    deferred.addCallback(quit, "SUCCESS")
    deferred.addErrback(failed)
    twisted.internet.reactor.callLater(60, quit, None, "TIMEOUT")
    msg = "SENDING CHANGES to buildbot master %s:%d on %s" % (
        server, port, hook)
    print " bzrbuildbot:", msg
    twisted.internet.reactor.run()

def post_commit(local_branch, master_branch, # branch is the master_branch
                old_revno, old_revid, new_revno, new_revid):
    if _installed_hook(master_branch) == COMMIT_VALUE:
        send_change(master_branch,
                     old_revid, old_revid, new_revno, new_revid, COMMIT_VALUE)

def post_push(result):
    if _installed_hook(result.target_branch) == PUSH_VALUE:
        send_change(result.target_branch,
                     result.old_revid, result.old_revid,
                     result.new_revno, result.new_revid, PUSH_VALUE)

def post_change_branch_tip(result):
    if _installed_hook(result.branch) == CHANGE_VALUE:
        send_change(result.branch,
                     result.old_revid, result.old_revid,
                     result.new_revno, result.new_revid, CHANGE_VALUE)
