import bzrlib.branch
import twisted.python.log
import twisted.internet.defer
import twisted.internet.reactor
import twisted.internet.task
import twisted.internet.threads

import buildbot.util
import buildbot.changes.base
import buildbot.changes.changes

import bzrbuildbot.change


class BzrPoller(buildbot.changes.base.ChangeSource,
                buildbot.util.ComparableMixin):

    compare_attrs = ['url']

    def __init__(self, url, poll_interval=10*60, blame_merge_author=False):
        # poll_interval is in seconds, so default poll_interval is 10 minutes
        # bzr+ssh://bazaar.launchpad.net/~launchpad-pqm/launchpad/devel/
        # works, lp:~launchpad-pqm/launchpad/devel/ doesn't.
        if url.startswith('lp:'):
            url = 'bzr+ssh://bazaar.launchpad.net/' + url[3:]
        self.url = url
        self.poll_interval = poll_interval
        self.loop = twisted.internet.task.LoopingCall(self.poll)
        self.blame_merge_author = blame_merge_author

    def startService(self):
        twisted.python.log.msg("BzrPoller(%s) starting" % self.url)
        buildbot.changes.base.ChangeSource.startService(self)
        twisted.internet.reactor.callWhenRunning(
            self.loop.start, self.poll_interval)
        for change in reversed(self.parent.changes):
            if change.branch == self.url:
                self.last_revision = change.revision
                break
        else:
            self.last_revision = None
        self.polling = False

    def stopService(self):
        twisted.python.log.msg("BzrPoller(%s) shutting down" % self.url)
        self.loop.stop()
        return buildbot.changes.base.ChangeSource.stopService(self)

    def describe(self):
        return "BzrPoller watching %s" % self.url

    @twisted.internet.defer.inlineCallbacks
    def poll(self):
        if self.polling: # this is called in a loop, and the loop might
            # conceivably overlap.
            return
        self.polling = True
        try:
            # On a big tree, even individual elements of the bzr commands can
            # take awhile.  So we just push the bzr work off to a thread.
            try:
                changes = yield twisted.internet.threads.deferToThread(
                    self.getRawChanges)
            except (SystemExit, KeyboardInterrupt):
                raise
            except:
                # we'll try again next poll.  Meanwhile, let's report.
                twisted.python.log.err()
            else:
                for change in changes:
                    change['branch'] = self.url
                    yield self.addChange(
                        buildbot.changes.changes.Change(**change))
                    self.last_revision = change['revision']
        finally:
            self.polling = False

    def getRawChanges(self):
        branch = bzrlib.branch.Branch.open_containing(self.url)[0]
        changes = []
        change = bzrbuildbot.change.generate(
            branch, blame_merge_author=self.blame_merge_author)
        if (self.last_revision is None or
            change['revision'] > self.last_revision):
            changes.append(change)
            if self.last_revision is not None:
                while self.last_revision + 1 < change['revision']:
                    change = bzrbuildbot.change.generate(
                        branch, new_revno=change['revision']-1,
                        blame_merge_author=self.blame_merge_author)
                    changes.append(change)
        changes.reverse()
        return changes

    def addChange(self, change):
        d = twisted.internet.defer.Deferred()
        def _add_change():
            d.callback(
                self.parent.addChange(change))
        twisted.internet.reactor.callLater(0, _add_change)
        return d
