import re

import twisted.python.log

import buildbot.process.buildstep
import buildbot.steps.shell


class TestObserver(buildbot.process.buildstep.LogLineObserver):
    # adds a summary log, and keeps track of test count as we go.

    _ignore_line = re.compile(
        r'( [\w\.\/\-]+( ?\([\w\.\/\-]+\))?|'
        r'\s*Running.*|'
        r'\d{4}\-\d{2}\-\d{2} \d{2}\:\d{2}\:\d{2} INFO.+|'
        r'\s*Set up .+|'
        r'\s*Tear down .*|'
        r'  Ran \d+ tests with .+)$').match

    #  Ran 223 tests with 0 failures and 0 errors in 3 minutes 0.483 seconds.
    _intermediate_summary = re.compile(
        r'  Ran (\d+) tests? with (\d+) failures? and (\d+) errors? '
        r'in .*?[\d\.]+ seconds?.').match

    #Total: 20302 tests, 0 failures, 0 errors in 141 minutes 5.392 seconds.
    _final_summary = re.compile(
        r'Total: (\d+) tests?, (\d+) failures?, (\d+) errors? '
        r'in .*?[\d\.]+ seconds?.').match

    test_count = 0
    failure_count = 0
    error_count = 0

    finished = False
    log = None

    def outLineReceived(self, line):
        if self.finished:
            return
        if self.log is None:
            self.log = self.step.addLog('summary')
        report = False
        match = self._intermediate_summary(line)
        if match:
            total, failures, errors = match.groups()
            self.test_count += int(total)
            self.failure_count += int(failures)
            self.error_count += int(errors)
        else:
            match = self._final_summary(line)
            if match:
                self.finished = True
                total, failures, errors = match.groups()
                self.test_count = int(total)
                self.failure_count = int(failures)
                self.error_count = int(errors)
                report = True
            else:
                report = not self._ignore_line(line)
        if report:
            self.log.addStdout(line + '\n')
        if match:
            twisted.python.log.msg(
                '%d tests, %d failures, %d errors' %
                (self.test_count, self.failure_count, self.error_count))
            self.step.setProgress('tests', self.test_count)
            self.step.step_status.setStatistic(
                'tests-total', self.test_count)
            self.step.step_status.setStatistic(
                'tests-failures', self.failure_count)
            self.step.step_status.setStatistic(
                'tests-errors', self.error_count)
            self.step.step_status.setText(self.step.describe(False))
        if self.finished:
            self.log.finish()


class Test(buildbot.steps.shell.ShellCommand):

    progressMetrics = ('output', 'tests')

    def __init__(self, **kwargs):
        if 'description' not in kwargs:
            kwargs['description'] = ['testing']
        if 'descriptionDone' not in kwargs:
            kwargs['descriptionDone'] = ['test']
        buildbot.steps.shell.ShellCommand.__init__(self, **kwargs)
        self.addLogObserver('stdio', TestObserver())

    def describe(self, done=False):
        description = buildbot.steps.shell.ShellCommand.describe(self, done)[:]
        if self.step_status.hasStatistic('tests-total'):
            total = self.step_status.getStatistic("tests-total", 0)
            failures = self.step_status.getStatistic("tests-failures", 0)
            errors = self.step_status.getStatistic("tests-errors", 0)
            passed = total - (failures + errors)

            if total:
                description.append('%d tests' % total)
            if passed:
                description.append('%d passed' % passed)
            if failures:
                description.append('%d failures' % failures)
            if errors:
                description.append('%d errors' % errors)
            if not done:
                description.extend(['(through', 'last', 'completed', 'layer)'])
        else:
            description.append("no test results")
        return description
